@@ -1,7 +1,8 @@
 use std::collections::HashMap;
 use std::hash::{DefaultHasher, Hash, Hasher};
-use std::io::Read;
+use std::io::{ErrorKind, Read};
 use std::path::{Path, PathBuf};
+use std::str::FromStr;
 use std::sync::atomic::{AtomicBool, Ordering};
 use std::sync::{Arc, Mutex};
 
@@ -9,15 +10,15 @@ use cargo::core::compiler::{unit_graph::UnitDep, unit_graph::UnitGraph, Executor
 use cargo::core::profiles::Profiles;
 use cargo::core::{FeatureValue, Package, PackageId, Target, TargetKind, Workspace};
 use cargo::ops::{self, CompileFilter, CompileOptions, FilterRule, LibRule};
-use cargo::util::command_prelude::{ArgMatches, ArgMatchesExt, CompileMode, ProfileChecking};
+use cargo::util::command_prelude::{ArgMatches, ArgMatchesExt, ProfileChecking, UserIntent};
 use cargo::util::interning::InternedString;
 use cargo::{CliResult, GlobalContext};
 
 use anyhow::Context as _;
+use cargo_platform::Platform;
 use cargo_util::paths::{copy, create_dir_all, open, read, read_bytes, write};
 use implib::def::ModuleDef;
 use implib::{Flavor, ImportLibrary, MachineType};
-use itertools::Itertools;
 use semver::Version;
 
 use crate::build_targets::BuildTargets;
@@ -29,7 +30,7 @@ use crate::target;
 fn build_include_file(
     ws: &Workspace,
     name: &str,
-    version: &Version,
+    version: Option<&Version>,
     root_output: &Path,
     root_path: &Path,
 ) -> anyhow::Result<()> {
@@ -43,15 +44,17 @@ fn build_include_file(
 
     // TODO: map the errors
     let mut config = cbindgen::Config::from_root_or_default(crate_path);
-    let warning = config.autogen_warning.unwrap_or_default();
-    let version_info = format!(
-        "\n#define {0}_MAJOR {1}\n#define {0}_MINOR {2}\n#define {0}_PATCH {3}\n",
-        name.to_uppercase().replace('-', "_"),
-        version.major,
-        version.minor,
-        version.patch
-    );
-    config.autogen_warning = Some(warning + &version_info);
+    if let Some(version) = version {
+        let warning = config.autogen_warning.unwrap_or_default();
+        let version_info = format!(
+            "\n#define {0}_MAJOR {1}\n#define {0}_MINOR {2}\n#define {0}_PATCH {3}\n",
+            name.to_uppercase().replace('-', "_"),
+            version.major,
+            version.minor,
+            version.patch
+        );
+        config.autogen_warning = Some(warning + &version_info);
+    }
     cbindgen::Builder::new()
         .with_crate(crate_path)
         .with_config(config)
@@ -112,7 +115,11 @@ fn patch_target(
     let manifest = pkg.manifest_mut();
     let targets = manifest.targets_mut();
 
-    let mut kinds = Vec::with_capacity(2);
+    let mut kinds = Vec::with_capacity(3);
+
+    if library_types.rlib {
+        kinds.push(CrateType::Rlib);
+    }
 
     if library_types.staticlib {
         kinds.push(CrateType::Staticlib);
@@ -303,7 +310,7 @@ impl FingerPrint {
         let mut f = open(self.path())?;
         let mut cache_str = String::new();
         f.read_to_string(&mut cache_str)?;
-        let cache = toml::de::from_str(&cache_str)?;
+        let cache = toml::from_str(&cache_str)?;
 
         Ok(cache)
     }
@@ -321,7 +328,7 @@ impl FingerPrint {
                 hash,
                 static_libs: self.static_libs.to_owned(),
             };
-            let buf = toml::ser::to_string(&cache)?;
+            let buf = toml::to_string(&cache)?;
             write(self.path(), buf)?;
         }
 
@@ -343,6 +350,7 @@ pub struct HeaderCApiConfig {
     pub subdirectory: String,
     pub generation: bool,
     pub enabled: bool,
+    pub emit_version_constants: bool,
 }
 
 #[derive(Debug, Hash)]
@@ -480,7 +488,7 @@ fn load_manifest_capi_config(
         .crate_name();
     let root_path = pkg.root().to_path_buf();
     let manifest_str = read(&root_path.join("Cargo.toml"))?;
-    let toml = manifest_str.parse::<toml::Value>()?;
+    let toml = manifest_str.parse::<toml::Table>()?;
 
     let capi = toml
         .get("package")
@@ -540,6 +548,11 @@ fn load_manifest_capi_config(
                 .and_then(|h| h.get("enabled"))
                 .map(|v| v.clone().try_into())
                 .unwrap_or(Ok(true))?,
+            emit_version_constants: header
+                .as_ref()
+                .and_then(|h| h.get("emit_version_constants"))
+                .map(|v| v.clone().try_into())
+                .unwrap_or(Ok(true))?,
         }
     } else {
         HeaderCApiConfig {
@@ -547,6 +560,7 @@ fn load_manifest_capi_config(
             subdirectory: String::from(name),
             generation: true,
             enabled: true,
+            emit_version_constants: true,
         }
     };
 
@@ -636,17 +650,38 @@ fn load_manifest_capi_config(
             });
         }
 
+        fn make_args(args: &str) -> impl Iterator<Item = String> + use<'_> {
+            args.split(' ')
+                .map(str::trim)
+                .filter(|s| !s.is_empty())
+                .map(str::to_string)
+        }
+
         import_library = library
             .get("import_library")
             .and_then(|v| v.as_bool())
             .unwrap_or(true);
+
         if let Some(args) = library.get("rustflags").and_then(|v| v.as_str()) {
+            rustflags.extend(make_args(args));
+        }
+
+        if let Some(args) = library.get("target").and_then(|v| v.as_table()) {
             let args = args
-                .split(' ')
-                .map(str::trim)
-                .filter(|s| !s.is_empty())
-                .map(str::to_string);
-            rustflags.extend(args);
+                .iter()
+                .filter_map(|(p, v)| {
+                    if Platform::from_str(p)
+                        .ok()
+                        .is_some_and(|p| p.matches(name, &rustc_target.cfg))
+                    {
+                        v.as_str()
+                    } else {
+                        None
+                    }
+                })
+                .flat_map(make_args);
+
+            rustflags.extend(args)
         }
     }
 
@@ -762,11 +797,11 @@ fn compile_options(
     gctx: &GlobalContext,
     args: &ArgMatches,
     profile: InternedString,
-    compile_mode: CompileMode,
+    compile_intent: UserIntent,
 ) -> anyhow::Result<CompileOptions> {
     use cargo::core::compiler::CompileKind;
     let mut compile_opts =
-        args.compile_options(gctx, compile_mode, Some(ws), ProfileChecking::Custom)?;
+        args.compile_options(gctx, compile_intent, Some(ws), ProfileChecking::Custom)?;
 
     compile_opts.build_config.requested_profile = profile;
 
@@ -848,7 +883,7 @@ impl Executor for Exec {
     }
 }
 
-use cargo::core::compiler::{unit_graph, UnitInterner};
+use cargo::core::compiler::{unit_graph, CompileMode, UnitInterner};
 use cargo::ops::create_bcx;
 
 fn set_deps_args(
@@ -927,7 +962,7 @@ fn compile_with_exec(
         .iter()
         .filter_map(|l| {
             let id = l.unit.pkg.package_id();
-            if let Some(ref m) = l.script_meta {
+            l.script_metas.iter().flatten().find_map(|m| {
                 if let Some(env) = r.extra_env.get(m) {
                     env.iter().find_map(|e| {
                         if e.0 == "OUT_DIR" {
@@ -939,9 +974,7 @@ fn compile_with_exec(
                 } else {
                     None
                 }
-            } else {
-                None
-            }
+            })
         })
         .collect();
 
@@ -1019,6 +1052,7 @@ fn deprecation_warnings(ws: &Workspace, args: &ArgMatches) -> anyhow::Result<()>
 pub struct LibraryTypes {
     pub staticlib: bool,
     pub cdylib: bool,
+    pub rlib: bool,
 }
 
 impl LibraryTypes {
@@ -1034,17 +1068,21 @@ impl LibraryTypes {
         Self {
             staticlib: true,
             cdylib: target.os != "none",
+            rlib: false,
         }
     }
 
-    fn from_args(target: &target::Target, args: &ArgMatches) -> Self {
-        match args.get_many::<String>("library-type") {
-            Some(library_types) => Self::from_library_types(target, library_types),
-            None => Self::from_target(target),
+    fn from_args(target: &target::Target, args: &ArgMatches, want_rlib: bool) -> Self {
+        Self {
+            rlib: want_rlib,
+            ..match args.get_many::<String>("library-type") {
+                Some(library_types) => Self::from_library_types(target, library_types),
+                None => Self::from_target(target),
+            }
         }
     }
 
-    pub(crate) fn from_library_types<S: AsRef<str>>(
+    fn from_library_types<S: AsRef<str>>(
         target: &target::Target,
         library_types: impl Iterator<Item = S>,
     ) -> Self {
@@ -1059,7 +1097,11 @@ impl LibraryTypes {
         // forcing a cdylib for musl is allowed here (see [`LibraryTypes::from_target`])
         cdylib &= target.os != "none";
 
-        Self { staticlib, cdylib }
+        Self {
+            staticlib,
+            cdylib,
+            rlib: false,
+        }
     }
 
     const fn only_staticlib(self) -> bool {
@@ -1081,7 +1123,6 @@ fn static_libraries(link_line: &str, rustc_target: &target::Target) -> String {
             }
             !s.is_empty()
         })
-        .unique()
         .map(|lib| {
             if rustc_target.env == "msvc" && lib.ends_with(".lib") {
                 return format!("-l{}", lib.trim_end_matches(".lib"));
@@ -1098,6 +1139,7 @@ pub fn cbuild(
     config: &GlobalContext,
     args: &ArgMatches,
     default_profile: &str,
+    tests: bool,
 ) -> anyhow::Result<(Vec<CPackage>, CompileOptions)> {
     deprecation_warnings(ws, args)?;
 
@@ -1109,13 +1151,13 @@ pub fn cbuild(
 
     let rustc_target = target::Target::new(Some(&target), is_target_overridden)?;
 
-    let library_types = LibraryTypes::from_args(&rustc_target, args);
+    let library_types = LibraryTypes::from_args(&rustc_target, args, tests);
 
     let profile = args.get_profile_name(default_profile, ProfileChecking::Custom)?;
 
     let profiles = Profiles::new(ws, profile)?;
 
-    let mut compile_opts = compile_options(ws, config, args, profile, CompileMode::Build)?;
+    let mut compile_opts = compile_options(ws, config, args, profile, UserIntent::Build)?;
 
     // TODO: there must be a simpler way to get the right path.
     let root_output = ws
@@ -1216,11 +1258,12 @@ pub fn cbuild(
 
             if capi_config.header.enabled {
                 let header_name = &capi_config.header.name;
+                let emit_version_constants = capi_config.header.emit_version_constants;
                 if capi_config.header.generation {
                     build_include_file(
                         ws,
                         header_name,
-                        &cpkg.version,
+                        emit_version_constants.then_some(&cpkg.version),
                         &root_output,
                         &cpkg.root_path,
                     )?;
@@ -1255,7 +1298,33 @@ pub fn cbuild(
                     from_build_targets.debug_info.as_ref(),
                     build_targets.debug_info.as_ref(),
                 ) {
-                    copy(from_debug_info, to_debug_info)?;
+                    if from_debug_info.exists() {
+                        create_dir_all(to_debug_info.parent().unwrap())?;
+                        if from_debug_info.is_dir() {
+                            let files =
+                                from_debug_info.read_dir()?.collect::<Result<Vec<_>, _>>()?;
+                            for f in files.iter() {
+                                let src = f.path();
+                                let file_name = src.strip_prefix(from_debug_info)?;
+                                let dst = to_debug_info.join(file_name);
+                                match std::fs::create_dir_all(
+                                    dst.parent().expect("Source path is not complete"),
+                                ) {
+                                    Ok(()) => Ok(()),
+                                    Err(v) => {
+                                        if v.kind() == ErrorKind::AlreadyExists {
+                                            Ok(())
+                                        } else {
+                                            Err(v)
+                                        }
+                                    }
+                                }?;
+                                copy(src, dst)?;
+                            }
+                        } else {
+                            copy(from_debug_info, to_debug_info)?;
+                        }
+                    }
                 }
             }
 
@@ -1285,7 +1354,7 @@ pub fn ctest(
 ) -> CliResult {
     compile_opts.build_config.requested_profile =
         args.get_profile_name("test", ProfileChecking::Custom)?;
-    compile_opts.build_config.mode = CompileMode::Test;
+    compile_opts.build_config.intent = UserIntent::Test;
 
     compile_opts.filter = ops::CompileFilter::new(
         LibRule::Default,   // compile the library, so the unit tests can be run filtered
@@ -1419,11 +1488,11 @@ mod tests {
         );
         assert_eq!(
             static_libraries(libs_msvc, &target_msvc),
-            "-lkernel32 -ladvapi32 -lntdll -luserenv -lws2_32 -lmsvcrt"
+            "-lkernel32 -ladvapi32 -lkernel32 -lntdll -luserenv -lws2_32 -lkernel32 -lws2_32 -lkernel32 -lmsvcrt"
         );
         assert_eq!(
             static_libraries(libs_mingw, &target_mingw),
-            "-lkernel32 -ladvapi32 -lntdll -luserenv -lws2_32"
+            "-lkernel32 -ladvapi32 -lkernel32 -lntdll -luserenv -lws2_32 -lkernel32 -lws2_32 -lkernel32"
         );
     }
 }
